/*----------------------------------------------------------------------
  AVR-Test- und Demo-programm V1.5
  Designed fr AVRCtrl V1.0

  (c) 2004 E.Hermanns, ed AT hermanns DOT net

  Absoluter AVR und AVR-GCC Erstversuch.

  Hardware: AVRCtrl V1.0, bestckt mit AT9028535, 8MHZ
  Compiler: avr-gcc V3.3.1
  
  Kompilieren mit "make" oder
  "/opt/cdk4avr/bin/avr-gcc -c -mmcu=at90s8535 -I. -g -Os -Wall -Wstrict-prototypes -std=gnu99 main.c -o main.o"

  Features:
  ---------
  - Menusystem mit Hilfetexten
  - Rc5Decoder im Polling Verfahren
  - Fernbedienbar uber (Tevion)-Fernbedienung (RC5)
  - LCD-Display Ansteuerung fr HD44780
  - Es koennen Displays bis vier Zeilen angesteuert werden.
  - Stoppuhr mit 1/1000 Sekunden Aufloesung, 9 Zwischenzeiten
    und Display-Update 1000 mal pro Sekunde. Die Fernbedienung
    ist allerdings maximal 1.8ms genau (= 0,0018 Sek).
  - Paralleltasks: die Stoppuhr kann im Hintergrund weiterlaufen, 
    whrend man andere Teile des Programms verwendet.
  - 6 von der Stoppuhr unabhaengige Zeitbasen zu
    500 sec, 1/1000 sec, 1/100 sec, 1/10 sec, 1 sec, 10 sec
    nach oben erweiterbar.

  NEU:
  ----
  - Vollstndiger DCF77 Decoder.

  Verwendete Ressourcen: 
  ----------------------
  - nur SIG_OUTPUT_COMPARE2, alle anderen Timer sind frei.
  - bislang keine explizite Prioritaetensteuerung
  - Flash zu 70 % benutzt, wobei ein grosser Teil davon
    zu Lasten der Hilfetexte geht.
       
  Geplant:
  --------
  - Setup-Routine fr andere Fernbedienungen
    Ablage der individuellen Tastencodes im EEPROM
  - Piezo-Summer fuer Alarmfunktionen
  - CountDown Uhr mit Alarm
  - Weckfunktion
  - Zwischenzeiten ww. als Total- oder
    Differenzzeit anzeigen
  
  History:
  --------
  20040214: V1.5rc2, V1.5 rel.cand. 2,
    auf LED 2 wird jetzt zur Kontrolle
    das DCF-Antennensignal angezeigt. 
  20040214: V1.5rc1, V1.5 rel.cand. 1,
    DCF-Dekoder voll funktionsfhig
  20040214: V1.5p5, V1.5 prerelease 5,
    DCF-Dekoder vollstaendig auf Tabellen
    verarbeitung umgestellt.
  20040214: V1.5p4, V1.5 prerelease 4,
    ca. 40 cases im DCF-Dekoder auf eine
    Tabelle umgestellt.
  20040213: V1.5p3, V1.5 prerelease 3,
    DCF77 Dekoder rechnet nun Uhrezeit
  20040212: V1.5p2, V1.5 prerelease 2, 
    Zeitstruktur vereinfacht, umgestellt auf
    uint8_t hh:mm:ss:hht
  20040212: V1.5p1, V1.5 prerelease 1, 
    DCF-77-Decoder hinzugefgt, 
    Einfache Fassung, dekodiert nur 0 und 1
  20040212: V1.4d, alle "bool" - Variablen
    konsquent von uint8_t auf bitfeld umgestellt.
    Es gibt jetzt ein globales Bitfeld, in dem 
    reine Flags angelegt werden. z.B: Bits.LcdIsBusy
  20040212: V1.4c, Sources neu organisiert
    Lcd, Dcf77 und Hardware -include Files erstellt 
    und #defines in eigene Dateien ausgelagert,
    Namen vereinheitlicht.
    Kommentare hinzugefgt.
  20040209: V1.4b, kleinerer Bugfix. Uebrig gebliebenes
    Debug-Statement entfernt.
  20040208: V1.4, Blink-Led "Stoppuhr laeuft" implementiert
    Flash jetzt bei 51% mit -Os
  20040208: V1.3, getch() geaendert. 
    Kann nun keine Fehlscans mehr zurueckgeben.
  20040208: V1.2, ClrScr-Routine gendert, 
    Kein "dazwischenschreiben" auf Display durch 
  Interrupt mehr moeglich.
  20040207: V1.1 Bugfixes.
    StHrMin[0] wurde durch einen strcpy_P auf DspBuf geloescht.
    Wurde in strncpy_P geaendert.
  20040207: V1.0 ins Netz gestellt
  ----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdio.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <string.h>
#include <avr/pgmspace.h>

#define HARDWARE "avrctrl.h"

#include "hd44780.h"
#include "dcf77.h"

#include HARDWARE

#ifndef TRUE
# define TRUE 0x01
# define FALSE 0x00
#endif

// Timer-Nachladewerte
#define _1000HZ_RELOAD		2
#define _100HZ_RELOAD		(1000/100)
#define _10HZ_RELOAD		(100/10)
#define _1HZ_RELOAD		(10/1)
#define _HZ10_RELOAD		10

/*------------------------------
  which Displaysize?
  ------------------------------*/

#define MY_DISP_LINES		2
#define MY_DISP_WIDTH		16
#define MY_DISP_BUF_LEN		(MY_DISP_LINES * MY_DISP_WIDTH)

// Definitionen Fernbedienung, passend fuer Tevion (aelteres Modell)

#define RC_RIGHT		54
#define RC_LEFT			53
#define RC_UP			32
#define RC_DOWN			33
#define RC_HELP			41
#define RC_OK			23
#define RC_MUTE			19
#define RC_KEY_M		55

#define getch			Rc5getch // use Remote Control as Keyboard

// Definitionen Stoppuhr

#define ST_FIRST_ZZ		0	// Erste Zwischenzeit
#define ST_ZZ			1	// andere Zwischenzeiten
#define ST_STOP			0
#define ST_RUN			1

// Funktionsmakros
#define EOn()		sbi(HW_HD_PORT, HW_HD_EN_BIT) // LCD - Enable ON
#define EOff()		cbi(HW_HD_PORT, HW_HD_EN_BIT) // LCD - Enable OFF
#define ROM_ADR(a)	((PGM_P) pgm_read_word(&(a))) // ROM - Adresse von Literal(a)

// Nur fuer 8MHZ geeignet, bei anderen Freq. entspr. angleichen
#define delay_1us() asm volatile("nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" \
				 "nop\n\t" )

/*----------------------------------------------------------------------
			   GLobal Variablen

  Alle die Variablen, die fuer den Compiler an irgendeiner "unsichtbaren"
  Stelle (ISR) benutzt werden (egal ob read oder write) muessen fuer den 
  Compiler als "volatile" markiert werden.
  ----------------------------------------------------------------------*/

/*-----------------------------------
  Allgemein verwendete Debug-Integer
  -----------------------------------*/
int DbgInt;

/*-----------------------------------
  typedefs
  -----------------------------------*/
struct PROGMEM RomMenu 
{
  PGM_P Hdr1;
  PGM_P Hdr2;
  PGM_P *Hlp;
};
typedef struct RomMenu ROM_MENU;

/*-----------------------------------
  Timer-Variablem
  -----------------------------------*/
volatile uint8_t  _1000Hz, _100Hz, _10Hz, _1Hz, _Hz10;

/*-----------------------------------
  Variablen Fernbedienung
  -----------------------------------*/
volatile uint8_t Rc5Code;
volatile uint8_t Rc5Unit;

volatile uint8_t Rc5Key;

/*-----------------------------------
  Variablen Lcd-Ansteuerung
  -----------------------------------*/
volatile char DspBuf[MY_DISP_BUF_LEN];
volatile char *DspPtr;

/*-----------------------------------
  Variablen Stoppuhr
  -----------------------------------*/

volatile uint8_t StStd[11];  // 0-255
volatile uint8_t StMin[11];  // 0-59
volatile uint8_t StSec[11];  // 0-59
volatile uint8_t StHun[11];  // 0-99, Hunderstel
volatile uint8_t StTsn[11];  // 0-9 Tausendstel
  
/*-----------------------------------
  DCF77
  -----------------------------------*/
volatile uint8_t DcfTime[8]; // Sekunde, Status, Minute, Stunde, Kalendertag, Wochentag, Monat, Jahr
volatile uint8_t DcfSample;
/*-----------------------------------
  Flags - alles was "bool" ist
  -----------------------------------*/
struct  
{
  // Dcf-Flags
  volatile unsigned int DcfSync 	: 1;
  volatile unsigned int DcfCurrent	: 1;
  volatile unsigned int DcfBit 		: 1;
  volatile unsigned int DcfLast 	: 1;
  // Fernbedienungs-Flags
  volatile unsigned int Rc5Mode		: 1;
  volatile unsigned int Rc5Current	: 1;
  volatile unsigned int Rc5Last		: 1;
  // Lcd-Flags
  volatile unsigned int LcdIsSync	: 1;
  volatile unsigned int LcdIsBusy	: 1;
  volatile unsigned int LcdCursor	: 1;
  // Stoppuhr-Flags
  volatile unsigned int StShowTime	: 1;
  volatile unsigned int StTimeChanged	: 1;
  volatile unsigned int StStopState	: 1;
} Bits;

/*-----------------------------------
  Menuetexte
  -----------------------------------*/
const char RomZ	 [] PROGMEM = "";	 

//				   "----+----1----+-"
const char HelpHelp0 [] PROGMEM = "Zurueck mit '<'" ;	
const char HelpHelp1 [] PROGMEM = "Vorwaerts: 'v'"	 ;
const char HelpHelp2 [] PROGMEM = "Rckwaerts: '^'"	 ;
const char HelpHelp3 [] PROGMEM = "Eingabe: 'Ok'"	 ;
const char HelpHelp4 [] PROGMEM = "Zurueck: '<'"	 ;
const char HelpHelp5 [] PROGMEM = "Cursor an/aus: 0";
const char HelpHelp6 [] PROGMEM = "    (c) 2004"	 ;
const char HelpHelp7 [] PROGMEM = "   E.Hermanns"	 ;

PGM_P HelpHelp [] PROGMEM = 
  {
    HelpHelp0, HelpHelp1, HelpHelp2, HelpHelp3, HelpHelp4, 
    HelpHelp5, HelpHelp6, HelpHelp7,
    RomZ
  };

//				   ----+----0----+-
const char StopHelp0 []	PROGMEM = "Starten/Stoppen"	; 
const char StopHelp1 []	PROGMEM = "mit >Ok<. Zwi-"	;
const char StopHelp2 []	PROGMEM = "schenzeiten 1-9"	;
const char StopHelp3 []	PROGMEM = "koennen mit 'M'"	; 
const char StopHelp4 []	PROGMEM = "genommen werden"	;
const char StopHelp5 []	PROGMEM = "Die Zwischen-"	;
const char StopHelp6 []	PROGMEM = "zeiten koennen"	;
const char StopHelp7 []	PROGMEM = "im Stoppuhrbe-"	;
const char StopHelp8 []	PROGMEM = "trieb jederzeit"	;
const char StopHelp9 []	PROGMEM = "mit den Tasten"	;
const char StopHelpa []	PROGMEM = "1-9 abgefragt"	;
const char StopHelpb []	PROGMEM = "werden, auch"	;
const char StopHelpc []	PROGMEM = "waehrend der"	;
const char StopHelpd []	PROGMEM = "Messung. Mit der"	;
const char StopHelpe []	PROGMEM = "Taste '<' kehrt"	;
const char StopHelpf []	PROGMEM = "man in das Menue"	; 
const char StopHelpg []	PROGMEM = "zurueck. Die Uhr"	; 
const char StopHelph []	PROGMEM = "kann in in der "	;
const char StopHelpi []	PROGMEM = "Zwischenzeit im"	;
const char StopHelpj []	PROGMEM = "Hintergrund "	;
const char StopHelpk []	PROGMEM = "weiter laufen."	;
const char StopHelpl []	PROGMEM = "Zurueck mit '<'"	;

PGM_P StopHelp [] PROGMEM =
  {
    StopHelp0, StopHelp1, StopHelp2, StopHelp3, StopHelp4, 
    StopHelp5, StopHelp6, StopHelp7, StopHelp8, StopHelp9, 
    StopHelpa, StopHelpb, StopHelpc, StopHelpd, StopHelpe,
    StopHelpf, StopHelpg, StopHelph, StopHelpi, StopHelpj,
    StopHelpk, StopHelpl,
    RomZ
  };

//				  "----+----1----+-"
const char Rc5Help0 []	PROGMEM = "Infrarot Codes"	;
const char Rc5Help1 []	PROGMEM = "abfragen. Nach"	;
const char Rc5Help2 []	PROGMEM = "Druck einer"		;
const char Rc5Help3 []	PROGMEM = "Taste wird der"	;
const char Rc5Help4 []	PROGMEM = "Tastencode und"	;
const char Rc5Help5 []	PROGMEM = "der Systemcode"	;
const char Rc5Help6 []	PROGMEM = "(Unit) angezeigt"	;
const char Rc5Help7 []	PROGMEM = "Wird vier mal"	;
const char Rc5Help8 []	PROGMEM = "hintereinander"	;
const char Rc5Help9 []	PROGMEM = "die gleiche"		;
const char Rc5Helpa []	PROGMEM = "Taste gedrueckt,"	;
const char Rc5Helpb []	PROGMEM = "wird wieder das"	;
const char Rc5Helpc []	PROGMEM = "Menue angezeigt."	;
const char Rc5Helpd []	PROGMEM = "Zurueck mit '<'"	;

PGM_P Rc5Help [] PROGMEM =
  {
    Rc5Help0, Rc5Help1, Rc5Help2, Rc5Help3, Rc5Help4,
    Rc5Help5, Rc5Help6, Rc5Help7, Rc5Help8, Rc5Help9,
    Rc5Helpa, Rc5Helpb, Rc5Helpc, Rc5Helpd, 
    RomZ
  };
//				  "----+----1----+-"

const char DcfHelp0 []	PROGMEM = "DCF-77 Dekoder"	;
const char DcfHelp1 []	PROGMEM = "bei der Arbeit"	;
const char DcfHelp2 []	PROGMEM = "zuschauen. Es"    	;
const char DcfHelp3 []	PROGMEM = "wird permanenet"	;
const char DcfHelp4 []	PROGMEM = "der aktuell ge-"	;
const char DcfHelp5 []	PROGMEM = "rechnete Zustand"	;
const char DcfHelp6 []	PROGMEM = "der Systenmzeit"	;
const char DcfHelp7 []	PROGMEM = "angezeigt"		;
const char DcfHelp8 []	PROGMEM = "Zurueck mit '<'"	;

PGM_P DcfHelp [] PROGMEM =
  {
    DcfHelp0, DcfHelp1, DcfHelp2, DcfHelp3, DcfHelp4,
    DcfHelp5, DcfHelp6, DcfHelp7, DcfHelp8, 
    RomZ
  };

//				   ----+----1----+-
const char MenuHdr10 [] PROGMEM = "***** MENU *****"	;	
const char MenuHdr20 [] PROGMEM = "Hilfe: >>>>>   ?"	;
const char MenuHdr11 [] PROGMEM = "Stoppuhr >>>  Ok"	;	
const char MenuHdr21 [] PROGMEM = "Stoppuhr Hilfe:?"	;
const char MenuHdr12 [] PROGMEM = "IR-Dekoder >	 Ok"	;	
const char MenuHdr22 [] PROGMEM = "Dekoder Hilfe: ?"	;	
const char MenuHdr13 [] PROGMEM = "DCF-Decoder > Ok"	;	
const char MenuHdr23 [] PROGMEM = "Dekoder Hilfe: ?"	;	

static ROM_MENU MainMenu [] PROGMEM =
  {
    //	 ----+----0----+-,   ----+----0----+-
    { MenuHdr10, MenuHdr20, (PGM_P *) HelpHelp },
    { MenuHdr11, MenuHdr21, (PGM_P *) StopHelp },
    { MenuHdr12, MenuHdr22, (PGM_P *) Rc5Help },
    { MenuHdr13, MenuHdr23, (PGM_P *) DcfHelp },
    { RomZ }
  };

//				     ----+----1----+-
const char Welcome1    [] PROGMEM = "AVR-Demo V1.5rc2"	;
const char Welcome2    [] PROGMEM = "(c) E.Hermanns"	;

const char Prompt      [] PROGMEM = "Taste druecken.."	;
const char Rc5UnitText [] PROGMEM = "Rc5Unit: ##"	;
const char Rc5CodeText [] PROGMEM = "Rc5Code: ## (#)"	;

const char WTage [] PROGMEM = "MoDiMiDoFrSaSo" ;	

void delay_us(uint16_t us)
{
  asm volatile (
		"mov	r30, %0"	"\n\t"	\
		"mov	r31, %1"	"\n\t"	\
		"adiw	r30, 0"		"\n\t"	\
		"breq	__delay_us_1"	"\n\t"	\
		"__delay_us_0:"		"\n\t"	\
		"nop"			"\n\t"	\
		"nop"			"\n\t"	\
		"nop"			"\n\t"	\
		"nop"			"\n\t"	\
		"sbiw	r30, 1"		"\n\t"	\
		"brne	__delay_us_0"	"\n\t"	\
		"__delay_us_1:"		"\n\t"	\
		: /* no outputs */		\
		: "r" (us & 0xff),		\
		"r" ((us >> 8) & 0xff)	\
		);
} // delay_us

/*-----------------------------------
  LCD-BusyFlag abfragen
  -----------------------------------*/
void LcdWaitBusy(void)
{
  sbi(HW_HD_PORT, HW_HD_RD_BIT);	// RD = 1
  cbi(HW_HD_PORT, HW_HD_RS_BIT);	// RS = 0, Instruction
  // upper nibble as input (temporary)	
  HW_HD_DDR = HW_HD_DDR & 0x07;
	
  do 
    {
      delay_1us();
      EOn();
      delay_1us(); //delay_1s();
    } while (bit_is_set(HW_HD_PIN, HW_HD_BUSY_BIT));
		
  // back to output port (default mode)
  HW_HD_DDR |= 0xF0;
} // LcdWaitBusy

/*-----------------------------------
  Zeichen auf LCD ausgeben, 4-Bit 
  -----------------------------------*/
void LcdPutch(const char const Data, char Type)
{ 
  Bits.LcdIsBusy = TRUE;
  LcdWaitBusy();
  
  cbi(HW_HD_PORT, HW_HD_RD_BIT);	// RD = 0
  
  if (Type == HD_DATA)
    sbi(HW_HD_PORT, HW_HD_RS_BIT); 
  
  EOn();
  HW_HD_PORT =	(uint8_t) ((HW_HD_PORT & 0x0F) | (Data & 0xF0));
  EOff();
	
  EOn();
  HW_HD_PORT =	(uint8_t) ((HW_HD_PORT & 0x0F) | (Data<<4));
  EOff();
  Bits.LcdIsBusy = FALSE;
} // lcd_putch 

/*-----------------------------------
  Bildschirm loeschen
  -----------------------------------*/
void LcdClrScr(void)
{
  memset ((char *) DspBuf, 0, sizeof (DspBuf));
  LcdPutch(HD_CLS, HD_CMD);  
  Bits.LcdIsSync = Bits.LcdIsBusy = TRUE;
  DspPtr = DspBuf;
  LcdPutch(HD_HOME, HD_CMD); // setzt Bits.LcdIsBusy automatisch zurueck
} // LcdPutch

/*-----------------------------------
  LCD-Cursor Positionieren
  -----------------------------------*/
void LcdGotoXy(uint8_t Col, uint8_t Line)
{
  const char HD_LineAddr[] = { 0x00, 0x40, 0x14, 0x54 };
  LcdPutch(HD_SET_DD_ADR + (Col % 40)	+ HD_LineAddr [Line % 4], HD_CMD);
} // LcdGotoXy

/*-----------------------------------
  Zeichen 4-Bitweise an LCD senden
  -----------------------------------*/
void LcdOutChar (void)
{  
  if (*DspPtr && *DspPtr != '#')  
    {
      if (!Bits.LcdIsSync)
	{
	  LcdGotoXy((DspPtr - DspBuf) % MY_DISP_WIDTH, (DspPtr - DspBuf) / MY_DISP_WIDTH);
	  Bits.LcdIsSync= TRUE;
	  return;
	}
		
      LcdPutch(*DspPtr, HD_DATA);
      *DspPtr=0;
    }
  else
    Bits.LcdIsSync=FALSE;
	
  if (DspPtr == (DspBuf + MY_DISP_BUF_LEN - 1))
    DspPtr = DspBuf;
  else
    {
      DspPtr++;
      if (!((DspPtr - DspBuf) % MY_DISP_WIDTH))
	Bits.LcdIsSync = FALSE;
    }
} // LcdOutChar

/*--------------------------------------------------
 *  Display initialization
 *--------------------------------------------------*/
void LcdInit(void)
{
  char Ch;
  static const char InitStr[] PROGMEM = {
    3,3,3,2,2,8,0,
    HD_CONTROL|HD_DISPLAY_ON|HD_CURSOR_OFF|HD_NO_BLINK,
    0,1,0,2,0,6 };
  uint8_t i;
  cbi (HW_HD_PORT, HW_HD_RS_BIT); // Instr. -> unset RS 
  delay_us(15000);
	
  for (i = 0; i < sizeof(InitStr); i++)
    {
      Ch = pgm_read_byte(&InitStr[i]);
      EOn(); 
      HW_HD_PORT = (HW_HD_PORT & 0x0F) | (Ch << 4); 
      EOff(); 
      delay_us(120);
      if (i == 0) delay_us(4000);
    };
		
  sbi (HW_HD_PORT, HW_HD_RS_BIT); // Data, -> set RS 
  Bits.LcdIsSync = FALSE;
} // LcdInit

/*-----------------------------------
  RC5 im Polling-Verfahren abfragen.
  
  Voraussetzung:
  Diese Routine muss alle 500s aufgerufen 
  werden. Da ein Bit immer ca. 1,8 ms lang 
  ist, wird bei langem low oder high der 
  Zhler "Period" bis mindestens drei 
  hochgezhlt, ansonsten ist er immer <= 2;

  Mehr nicht, keine Interrupts
  -----------------------------------*/
void Rc5Scan (void)
{
  static uint8_t BitNo = 0;
  static uint8_t Period = 0;
  static uint8_t Value = 0;
  uint8_t State;

  Bits.Rc5Current = bit_is_clear(PIND, PIND2); // IR-Sensor ist low-aktiv, deshalb "bit_is_clear"

  if (Bits.Rc5Last == Bits.Rc5Current) {
    if (Period == 4) {		// timeout, vergisses (wird in 99% aller Flle so sein */
      Period = BitNo = Value = 0;
      Bits.Rc5Mode = 1;
    } else Period++;
    return;
  }  

  // es gab eine Flanke
  State = 0;

  if (Period > 2) 		State = 1;  // war lange periode
  if (Bits.Rc5Last == TRUE)	State += 2; // war high

  switch (State) {
  case 0:			// war kurz und low
    BitNo += Bits.Rc5Mode;
    Value <<= Bits.Rc5Mode;
    Value += Bits.Rc5Mode;
    break;
    
  case 1:			// war lang und low
    Bits.Rc5Mode = TRUE;
    BitNo ++;
    Value <<=  1;
    Value ++;
    break;
    
  case 2:			// war kurz und high 
    if (!Bits.Rc5Mode == TRUE) {
      BitNo += 1;
      Value <<=	 1;
    }
    break;
      
  case 3:			// war lang und high
    Bits.Rc5Mode = FALSE;
    BitNo ++;
    Value <<= 1;
    break;
  }
 
  switch (BitNo) {
  case 2:			// startbits
    Value = 0;
    break;
      
  case 3:			// toggle-bit, verarbeiten, wenn erwnscht
    if (Value)
      {
	Rc5Code = 0x80;
	sbi(HW_OUT_PORT, HW_TOGGLE);
      }
    else
      {
	Rc5Code = 0;
	cbi(HW_OUT_PORT, HW_TOGGLE);
      }
    Value = 0;
    break;
      
  case 8:
    if (Value) {
      Rc5Unit = Value; // 5 bit
      Value = 0;
    }
    break;
      
  case 14:	  // bit 7 -> Toggle (0x80)
    // Bit 6 -> Rc5Code gescannt (0x40)
    // Bits 0-5 -> Scancode
    Rc5Code |= (Value | 0x40) ; 
    BitNo = 0;
    Value = 0;
    Bits.Rc5Mode = 1;
    break;
  }
  Bits.Rc5Last = Bits.Rc5Current;
  Period = 1;
} // Poll RC5

uint8_t Rc5getch(void)
{
  static uint8_t Toggle = 0;
  
  while (!((uint8_t) (Rc5Code & 0x40) &&    // successfull scan 
	   (uint8_t) (Rc5Code & 0x80) != Toggle) ) // toggle
    asm (" nop\n");
				
  Rc5Key = (uint8_t) (Rc5Code & 0x3f); // bits 0-5
  Toggle = (uint8_t) Rc5Code & 0x80;
  Rc5Code = 0;
  return Rc5Key;	
} // Rc5Scan

/*-----------------------------------
  DCF77-Signal Scanner
  -----------------------------------*/
void Dcf77Scan(void)
{
  static uint8_t BitNo, Period, Vx, Value, OldVal;
  uint8_t State;

  /*----------------------------------------------------------------------
    Dieses Array enthaelt die Grenzwerte fuer alle Flanken im 
    DCF77-Protokoll, die sich laut DCF_SAMPLE in dcf77.h ergeben. Das Array 
    wird einfach abgegrast und mit "Period" verglichen, so dass der Zaehler 
    bei dem Element "einrastet", das dem aktuellen Vorgang entspricht.
    ----------------------------------------------------------------------*/
  static const uint8_t DcfLimits[] PROGMEM = {
    DCF_BIT_OFF_COUNT_MIN,
    DCF_BIT_OFF_COUNT_MAX,
    DCF_BIT_ON_COUNT_MIN,
    DCF_BIT_ON_COUNT_MAX,
    DCF_LOW_COUNT_MIN,
    DCF_LOW_COUNT_MAX,
    DCF_59_LOW_COUNT_MIN,
    DCF_59_LOW_COUNT_MAX  };

  /*----------------------------------------------------------------------
    dieses Array enthaelt die Wertigkeiten fuer jedes Bit im DCF77-Ptotokoll,
    beginnend mit Bit 15, dem ersten Informations-Bit. 
    Zahlen > 100 sind Paritaets-Informationen.
    ----------------------------------------------------------------------*/
  static const uint8_t DcfValues[] PROGMEM = { 
    1,2,4,8,16,32,		// Bits 15-20: Reserveantenne (15), 
				//             Wechsel SZ/WZ anstehend (16), 	
				//             Sommerzeit (17)
				//             Winterzeit (18)
				//             Schaltsekunde anstehend (19)
				//             Startbit Zeitinformation (20)
    1,2,4,8,10,20,40,		// Bits 21-27: Minute
    107,			// Bit     28: Parity fuer die letzten 7 Bits
    1,2,4,8,10,20,		// Bits 29-34: Stunde
    106,			// Bit     35: Parity fuer die letzten 6 Bits
    1,2,4,8,10,20,		// Bits 36-41: Kalendertag
    1,2,4,			// Bits 42-44: Wochentag
    1,2,4,8,10,			// Bits 45-49: Monat
    1,2,4,8,10,20,40,80,	// Bits 50-57: Jahr
    122				// Bit     58: Parity fuer die letzten 22 Bits
  };
  
#define DCF_RESET   (Period = 1, BitNo = Bits.DcfSync = Vx = 0)

  Bits.DcfCurrent = (bit_is_set(HW_DCF_IN_PORT, HW_DCF_PIN)) ? 1 : 0;

  if (Bits.DcfLast == Bits.DcfCurrent) {
      if (Period == DCF_59_LOW_COUNT_MAX)	// timeout, vergisses 
	DCF_RESET;
      else 
	Period++;

      if (!(Period > DCF_LOW_COUNT_MAX && BitNo == 58)) 
	return;
    }  

  // entweder gab es eine Flanke oder BitNo 59 

  for (State = 0; State < sizeof (DcfLimits) - 2; State += 2)
    {
      if (Period >= pgm_read_byte(&(DcfLimits[State])) &&
	  Period <= pgm_read_byte(&(DcfLimits[State+1])))
	break;
    }

  if (State != 6)		// 59te Sekunde
    State += (Bits.DcfLast <<4); 

  Period = 1;
  Bits.DcfLast = Bits.DcfCurrent;

  if (State == 4) return;	// gueltige Low-Zeit

  if (BitNo < 59)
    BitNo++;
  else
    BitNo = 0;

  switch (State) 
    {
    case 0x10: Bits.DcfBit = 0; break;  // war kurz und high:	    0-bit
    case 0x12: Bits.DcfBit = 1; break;  // war lang und high:	    1-bit
    case 0x06: BitNo = 59;      break;  // war Doppelpause und low: 59. Sekunde
    default:   DCF_RESET;       return;	// war ungueltig
    }

  // bis hier wurde nur der Zustand des Dekoders gerechnet
  // ab hier wird die uhrzeit errechnet

  /*----------------------------------------------------------------------
    Das Verfahren ist simpel:

    Jedes Bit zaehlt die Sekunde (DcfTime[0]) hoch.
    Bits 0-14 werden nicht weiter verarbeitet.
    Bei Bit Nr. 15 wird der Zeitindex fr das DcfTime[] array auf 1 gesetzt.
    Das entprechende array-Element wird nun immer um den Wert erhht, der sich aus 
    Bits.DcfBit * DcfValues[BitNr] errechnet. Ist der gelesene DcfValues-Wert 
    kleiner als der vorhergehende, wird zuvor der DcfTime-Array-Index um eins erhht.
    Das ist alles.

    Werte ber hundert werden uebrigens nicht addiert, sondern zeigen an, das inklusive 
    dieser Stelle eine even Parity ber die vorhergehenden (Value-100) Bits vorliegen muss.
    ----------------------------------------------------------------------*/
  DcfTime[0] = BitNo;
  
  if (BitNo < 15 || BitNo > 58) {
    if (BitNo == 0) {
      memset ((char *) &DcfTime[1], 0, sizeof (DcfTime) - sizeof (DcfTime[0]));
      OldVal = 0;
    }
    return;
  }

  if (BitNo == 15) Vx = 1;	// Value-Index erhoehen

  Value = pgm_read_byte(&(DcfValues[BitNo - 15]));
  if (Value > 100) return;	// TODO: Parityberechnung
  if (Value < OldVal) Vx++;	// Nchstes Datenelement
  if (Bits.DcfBit == 1) DcfTime[Vx] += Value;
  OldVal = Value;
} // Dcf77Scan

/*-----------------------------------
  Die Stoppuhr ... schnelle Version
  worst Case: 189,5 us vom Aufruf bis 
  zur Rueckkehr 
  -----------------------------------*/
void StIncTime(void)
{
  uint8_t Wrk;
  /*----------------------------------------------------------------------
    Aufbau der Zeit-Struktur:
    Element [0] enthaelt die derzeit gestoppte Zeit.
    Elemente [1-9] enthalten die Zwischenzeiten
    Ab Element 10 koennen weitere Zeiten (z.B. DCF-Zeit) gespeichert werden.
    
    Hunderstel und Tausendstel wurden wurden in zwei einzelne 8-Bit Variablen
    gestellt. Das kostet genauso viel Speicher wie eine int16 Variable fr 
    die Sekundenbruchteile 0,001 bis 0,999 kosten wrde, ist aber schneller 
    in der Verarbeitung.
    ----------------------------------------------------------------------*/		

  if (StTsn[0] < 9) StTsn[0]++;
  else {
    StTsn[0] = 0;
    if (StHun[0] < 99) StHun[0]++;
    else {
      StHun[0] = 0;
      if (StSec[0] < 59) StSec[0]++;
      else {
	StSec[0] = 0;
	if (StMin[0] < 59) StMin[0]++;
	else {
	  StMin[0] = 0;
	  StStd[0]++;
	  // ----+----1----+-
	  // 000hhh:mm:ss,ttt
	  if (Bits.StShowTime) {
	    Wrk = StStd[0] % 10;
	    DspBuf[5] = Wrk + '0';
	    if (!Wrk) { //  Ueberlauf
	      Wrk = (StStd[0] / 10 ) % 10;
	      DspBuf[4] = Wrk + '0';
	      if (!Wrk) { //  Ueberlauf
		DspBuf[3] = (StStd[0]/100) + '0';
	      } // Hunderter Stunden dargestellt
	    } // Zehner Stunden dargestellt
	  } // Einer Stunden dargestellt
	} // if (Bits.StShowTime) fuer Stunden
	if (Bits.StShowTime) {
	  Wrk = StMin[0] % 10;
	  DspBuf[8] = Wrk + '0';
	  if (!Wrk) // Ueberlauf
	    DspBuf[7] = (StMin[0]/10) + '0';
	}  // if (Bits.StShowTime) fuer Minuten
      }
      if (Bits.StShowTime) {
	Wrk = StSec[0] % 10;
	DspBuf[11] = Wrk + '0';
	if (!Wrk) // Ueberlauf
	  DspBuf[10] = (StSec[0]/10) + '0';
      } // if (Bits.StShowTime) fuer Sekunden
    }
    if (Bits.StShowTime) {
      Wrk = StHun[0] % 10;
      DspBuf[14] = Wrk + '0';
      if (!Wrk) // Ueberlauf
	DspBuf[13] = (StHun[0]/10) + '0';
    } // if (Bits.StShowTime) fuer Hunderstel
  }
  if (Bits.StShowTime) {
    DspBuf[15] = StTsn[0] + '0';
  } // if (Bits.StShowTime) fuer Tausendstel
} // StIncTime, schnelle Version
  

/*-----------------------------------
  Zeitspeicher (n) anzeigen
  -----------------------------------*/
void StDspTime(uint8_t Nbr, char Mode)
{		    
  // Display-Positions fr:'   HHH:MM:SS:ttt'
  // >>>>>>>>>>>>>>>>>>>>>:'0123456789ABCDEF' 
  //	code char DspPos[]={3,4,5,7,8,10,11,13,14,15}; 

  char *Ch;
  if (Nbr)
    Ch = (char *) DspBuf + MY_DISP_WIDTH;
  else
    Ch = (char *) DspBuf;
  // clrtoeol ...
  memset (Ch, ' ', MY_DISP_WIDTH * sizeof(char));

  *Ch = Mode;
  Ch++;
  if (Nbr < 10)
    *Ch = Nbr + '0'; 
  
  Ch += 2;

  if (StStd[Nbr] > 99) 
    *Ch	  = (StStd[Nbr] / 100) + '0';	
  else 		  
    *Ch = ' ';
  Ch++;

  if (StStd[Nbr] > 9) 
    *Ch = ((StStd[Nbr] / 10) % 10) + '0'; 
  else
    *Ch = ' ';
  
  *(++Ch) = (StStd[Nbr] % 10) + '0'; 
  *(++Ch) = ':'; 
  *(++Ch) = (StMin[Nbr] / 10) + '0';								
  *(++Ch) = (StMin[Nbr] % 10) + '0';		
  *(++Ch) = ':';
  *(++Ch) = (StSec[Nbr] / 10) + '0';				
  *(++Ch) = (StSec[Nbr] % 10) + '0';					
  *(++Ch) = ',';
  *(++Ch) = (StHun[Nbr] % 10) + '0';		
  *(++Ch) = (StHun[Nbr] / 10) + '0';		
  *(++Ch) = (StTsn[Nbr] % 10)  + '0';

} // StDspTime

/*-----------------------------------
  Zeitspeicher loeschen
  -----------------------------------*/
void StClrTime(uint8_t StartZz)
{
  // ClrMode: FIRST_ZZ nur erste Zeit, ZZ Zwischenzeiten [1-9]
  uint8_t n;
  for (n = StartZz; n < 10; n++)
    {
      StStd[n] = StMin[n] = StSec[n] = StHun[n] = StTsn[n] = 0;
      if (StartZz == ST_FIRST_ZZ) return;
    }
} // StClrTime

/*-----------------------------------
  Zeit speichern
  -----------------------------------*/
void StStoreTime(uint8_t Zz)
{
  do
    {
      Bits.StTimeChanged = FALSE; // Check for Timer-Interrupt
      StStd[Zz] = StStd[0];
      StMin[Zz] = StMin[0];
      StSec[Zz] = StSec[0];
      StHun[Zz] = StHun[0];
      StTsn[Zz] = StTsn[0];
    } while (Bits.StTimeChanged);
} // StStoreTime

/*-----------------------------------
  Stoppuhr-Anzeige und -Interaktion
  -----------------------------------*/
void StStop(void)
{
  static uint8_t Zz = 0;	// Zw.Zeit-Index
	
  LcdClrScr();
  
  StDspTime(0, (Bits.StStopState) ? 'R' : 'S');

  if (Zz)
    StDspTime(Zz, 'Z');
		
  Bits.StShowTime = TRUE;
	
  while (TRUE)
    {
      Rc5Key = getch();
		
      switch (Rc5Key)
	{
	case RC_LEFT: // Menu wieder anzeigen
	  Bits.StShowTime = FALSE;
	  return;
	
	case RC_KEY_M:	// ZZ speichern
	  if (Bits.StStopState == TRUE )
	    if (Zz < 9)
	      {
		StStoreTime(++Zz);
		StDspTime(Zz, 'Z');
	      }
	  break;

	case RC_OK:
	  if (Bits.StStopState == ST_RUN)
	    StDspTime(0, 'S');
	  else // start clock
	    {
	      StClrTime(ST_FIRST_ZZ); /* zeit[0] loeschen */
	      StClrTime(ST_ZZ);	/* alle anderen zeiten loeschen */
	      // Clear line 2
	      memset ((char *) DspBuf + MY_DISP_WIDTH, ' ', MY_DISP_WIDTH * sizeof(char));
	      StDspTime(0, 'R');
	      Zz=0;
	    }
	  Bits.StStopState = ~Bits.StStopState; // Toggle runmode
	  break;
	
	case RC_MUTE: // Pause
	  Bits.StStopState = ST_STOP;
	  StDspTime(0, 'P');
	  while (getch() != RC_MUTE)
	    ;
	  Bits.StStopState = ST_RUN;
	  StDspTime(0, 'R');
	  break;

	case 1: 
	case 2: 
	case 3: 
	case 4: 
	case 5: 
	case 6: 
	case 7: 
	case 8: 
	case 9: 
	  Bits.StShowTime = FALSE; 
	  StDspTime(Rc5Key, 'D');
	  Bits.StShowTime = TRUE; 
	  break;			
	} // switch
    } // while TRUE
} // StStop

/*-----------------------------------
  Zeit-Interrupt, Timer2, Compare 
  -----------------------------------*/
SIGNAL(SIG_OUTPUT_COMPARE2)
{
  // wird alle 500 Sec aufgerufen
  cli();			// disable interrupts
  Rc5Scan();			// check RC5-Ir
  
  if (--_1000Hz)
    {
      sei();
      return;
    }

  /*----- diese Sektion wird alle 1/1000 sec ausgefhrt -----*/
  _1000Hz = _1000HZ_RELOAD;
  
  if (Bits.StStopState == ST_RUN)	StIncTime();
  if (!Bits.LcdIsBusy) 			LcdOutChar();

  if (DcfSample == DCF_SAMPLE)
    {
      Dcf77Scan();
      DcfSample = 1;
    }
  else
    DcfSample++;

  if (--_100Hz) 
    {
      sei();
      return;
    }

  /*----- diese Sektion wird alle 1/100 sec ausgefhrt -----*/
  _100Hz = _100HZ_RELOAD;	

  // Blink - Led  "Messung laeuft" 
  if (Bits.StStopState == ST_RUN && _10Hz == 2) sbi (HW_OUT_PORT, HW_MEASURE);

  // show DCF-Signal 
  if (bit_is_set(HW_DCF_IN_PORT, HW_DCF_PIN))
    sbi (HW_OUT_PORT, HW_DCF_SIG);
  else
    cbi (HW_OUT_PORT, HW_DCF_SIG);
	

  if (--_10Hz)
    {
      sei();
      return;
    }

  /*----- diese Sektion wird alle 1/10 sec ausgefhrt -----*/
  _10Hz = _10HZ_RELOAD;

  cbi (HW_OUT_PORT, HW_MEASURE);

  // kleine "alive" - led
  if (_1Hz == 6) sbi(HW_OUT_PORT, HW_ALIVE);
  if (_1Hz == 5) cbi(HW_OUT_PORT, HW_ALIVE);
	
  // wer eine Variable Wait_10Hz definiert, kann diese im Programm
  // setzen, um z.B. abzufragen, ob die Anzahl 10tel Sekunden bereits verstrichen ist...
  //	if (Wait_10Hz)
  //		Wait_10Hz--;
	
  if (--_1Hz) 
    {
      sei();
      return;
    }
	
  /*----- diese Sektion wird alle 1/1 sec ausgefhrt -----*/
  _1Hz = _1HZ_RELOAD;
  
  if (--_Hz10)
    {
      sei();
      return;
    }
	
  /*----- diese Sektion wird alle 10/1 sec ausgefhrt -----*/
  _Hz10 = _HZ10_RELOAD;
  
  sei();
} // SIGNAL(SIG_OUTPUT_COMPARE2)
  
/*-----------------------------------
  System initialisieren
  -----------------------------------*/
void InitSystem(void)
{ 
  // Clear timer on compare match 
  TCCR2 = (1<<CTC2);
  
  // PollFreq = 500Sec = 2000Hz = 8MHz/(32*125)
  TCCR2 |= (1<<CS21)|(1<<CS20);	 // presacler -> 32 -> CS21 and CS20
  OCR2 =   124;			 // Set Output Compare Value to 125
  TIFR =   1<<OCF2;		 // Clear OCF2,	 Clear pending interrupts 
  TIMSK =  1<<OCIE2;		 // Enable Timer2 Output Compare Match Interrupt
  HW_OUT_DDR  =  0xFF;		 // Set Port B as output
  HW_HD_DDR = 0xFF;		
  HW_DCF_DDR = HW_DCF_DDR & (0xFF ^ (1<<HW_DCF_PIN));  // pin3 (0x08) nicht v. Display benutzt
  
  _1000Hz = _1000HZ_RELOAD; // 1000 Hz
  _100Hz =  _100HZ_RELOAD;	// 100 Hz 
  _10Hz =   _10HZ_RELOAD;	// 10 Hz 
  _1Hz =    _1HZ_RELOAD;	// 1 Hz
  _Hz10 =   _HZ10_RELOAD;	// 10 Hz
  DcfSample = 0;		// 25 ms
  DspPtr = DspBuf;
} // InitSystem

/*-----------------------------------
  RC5-Scancodes anzeigen
  -----------------------------------*/
void Rc5Display(void)
{
  unsigned char LastKey = 255;
  uint8_t Count = 0;
  char *ch;
	
  LcdClrScr();
  strncpy_P((char *) DspBuf, Prompt, MY_DISP_WIDTH);
	
  while (TRUE)
    {  
      Rc5Key = getch();

      if (Rc5Key == LastKey)
	{
	  if (Count == 3) return;
	  Count++;
	}
      else
	{
	  Count = 1;
	  LastKey = Rc5Key;
	}
	
      LcdClrScr();
      // lcd_putstr_P (Disp);
      strcpy_P((char *) DspBuf, Rc5UnitText);
      for (ch=(char *) DspBuf; *ch != '#'; ch++)
	;
      *ch =  Rc5Unit / 10 + '0'; ch++;
      *ch =  Rc5Unit % 10 + '0'; 
      // LcdPutstr (Disp);

      ch = (char *) DspBuf + MY_DISP_WIDTH;
      strcpy_P(ch, Rc5CodeText);
      while (*ch != '#')   ch++;
      *ch = Rc5Key / 10 + '0';	ch++;
      *ch = Rc5Key % 10 + '0';	ch++;
      while (*ch != '#')	ch++;
      *ch = Count%10 + '0';
    }
} // Rc5Display

/*-----------------------------------
  DCF-Dekoder Zustand anzeigen
  -----------------------------------*/
void DcfDisplay(void)
{
  uint8_t SCount = 0;

  while (TRUE)
    {  
      LcdClrScr();
      
      while (!Rc5Code)
	{
	  if (_10Hz == 1)
	    {
	     if (DcfTime[5])	// Wochentag
	       { 
		 DspBuf[0] = pgm_read_byte(&WTage[(DcfTime[5]-1) * 2]);
		 DspBuf[1] = pgm_read_byte(&WTage[((DcfTime[5]-1) * 2)+1]);
	       }

	      DspBuf[ 3] = (DcfTime[4] / 10) + '0';
	      DspBuf[ 4] = (DcfTime[4] % 10) + '0';
	      DspBuf[ 5] = '.';
	      DspBuf[ 6] = (DcfTime[6] / 10) + '0';
	      DspBuf[ 7] = (DcfTime[6] % 10) + '0';
	      DspBuf[ 8] = '.';
	      DspBuf[ 9] = (DcfTime[7] / 10) + '0';
	      DspBuf[10] = (DcfTime[7] % 10) + '0';

	      StStd[10] = DcfTime[3];
	      StMin[10] = DcfTime[2];
	      StSec[10] = DcfTime[0];


	      StDspTime(10, 'D');
	      SCount = 0;
	    }
	} // while !Rc5Code

      Rc5Key = getch();
      if (Rc5Key == RC_LEFT)
	return;

    } // while true
} // DcfDisplay

/*-----------------------------------
  Hilfetext anzeigen
  -----------------------------------*/
void Help(const PGM_P *HlpTxt)
{
  uint8_t h = 0;

  while (TRUE)
    {
      LcdClrScr();
      strncpy_P((char *) DspBuf, ROM_ADR(HlpTxt[h]), MY_DISP_WIDTH); // kein strcpy_P verwenden
      if (pgm_read_byte(ROM_ADR(HlpTxt[h+1])))
	strncpy_P((char *) DspBuf + MY_DISP_WIDTH, ROM_ADR(HlpTxt[h+1]), MY_DISP_WIDTH);
  
      Rc5Key = getch();
	  
      switch (Rc5Key)
	{
	case RC_UP:		 
	  if (h > 0) h--; 
	  break;
	case RC_DOWN:	 
	  if (pgm_read_byte(ROM_ADR(HlpTxt[h+1]))) h++; 
	  break;
	case RC_LEFT: 
	  return;
	}
    }
} // Help

/*-----------------------------------
  Hauptroutine
  -----------------------------------*/
int main (void)
{
  uint8_t h;
  InitSystem();
  LcdInit();
  Bits.Rc5Mode = TRUE;		// manchester, high = TRUE
  sei();			// enable interrupts

  // Welcome
  LcdClrScr();
  strncpy_P((char *) DspBuf, Welcome1, MY_DISP_WIDTH);
  strncpy_P((char *) DspBuf + MY_DISP_WIDTH, Welcome2, MY_DISP_WIDTH);
   
  /*
    wait 4 sec. 
    Wenn die Warteformel "(_Hz10 + Offset) % 10" ist,
    dann ist die Wartezeit (10 - Offset) Sekunden. 
    Ist ja wohl Logisch.
  */
 
  h = (_Hz10 + 6) % 10; // lease "h"
  while (h != _Hz10) 
    asm (" nop\n");


  h = 0;
  while (TRUE)
    {
      LcdClrScr();
      // BugFix: strncpy_P statt strcpy_P

      strncpy_P((char *) DspBuf, ROM_ADR(MainMenu[h].Hdr1), MY_DISP_WIDTH);
      strncpy_P((char *) DspBuf + MY_DISP_WIDTH, ROM_ADR(MainMenu[h].Hdr2), MY_DISP_WIDTH);

      Rc5Key = getch();
      switch (Rc5Key)
	{
	case RC_UP:		 
	  if (h > 0) h--; 
	  break;
	case RC_DOWN:	 
	  if (pgm_read_byte(ROM_ADR(MainMenu[h+1].Hdr1)))
	    h++; 
	  break;
	case RC_HELP: 
	  // Ch_P = (PGM_P) pgm_read_word (*Ch_P);
	  Help ((PGM_P *) ROM_ADR (MainMenu[h].Hlp));
	  break;
	case RC_OK: 
	  if (h == 1) StStop();
	  else if (h == 2) Rc5Display();
	  else if (h == 3) DcfDisplay();
	  break;
	case 0: 
	  if (Bits.LcdCursor)
	    {
	      LcdPutch (HD_CONTROL|HD_DISPLAY_OFF, HD_CMD);
	      LcdPutch (HD_CONTROL|HD_DISPLAY_ON|HD_CURSOR_OFF, HD_CMD);
	      Bits.LcdCursor = FALSE;
	    }
	  else
	    {
	      LcdPutch (HD_CONTROL|HD_DISPLAY_OFF, HD_CMD);
	      LcdPutch (HD_CONTROL|HD_DISPLAY_ON|HD_CURSOR_ON|HD_BLINK, HD_CMD);
	      Bits.LcdCursor = TRUE;
	    }
	  break;

	case RC_KEY_M:		/* debug */
	  DspBuf[0] = DbgInt/10 + '0';
	  DspBuf[1] = DbgInt%10 + '0';
	  if ( bit_is_set(HW_DCF_IN_PORT, HW_DCF_PIN))
	    DspBuf[3] = '1';
	  else
	    DspBuf[3] = '0';
	  getch();		/* wait for key */
	  break;
	}
    }
} // main
